import functools
import gc

import telebot
from sqlalchemy.future import Engine
from telebot.async_telebot import AsyncTeleBot

from .. import database, encryption, generate_password
from ..account_checks import (
    check_account,
    check_account_name,
    check_login,
    check_password,
)
from ..account_parsing import accounts_to_json, json_to_accounts
from ..decrypted_account import DecryptedAccount
from . import markups
from .helper_functions import (
    base_handler,
    delete_message,
    escape,
    get_state,
    register_state,
    send_deleteable_message,
    send_tmp_message,
)

Message = telebot.types.Message


async def get_accounts(
    bot: AsyncTeleBot,
    engine: Engine,
    mes: Message,
) -> None:
    await base_handler(bot, mes)
    accounts = database.get.get_accounts(
        engine,
        mes.from_user.id,
        to_sort=True,
    )
    if not accounts:
        return await send_tmp_message(bot, mes.chat.id, "У вас нет аккаунтов")

    # Make accounts copyable and escape special chars
    accounts = [f"`{account}`" for account in accounts]
    await send_deleteable_message(
        bot,
        mes.chat.id,
        "Ваши аккаунты:\n"
        + "\n".join(accounts)
        + "\nНажмите на название, чтобы скопировать\n"
        f"Всего {len(accounts)}",
    )


async def delete_all(bot: AsyncTeleBot, engine: Engine, mes: Message) -> None:
    await base_handler(bot, mes)
    bot_mes = await bot.send_message(
        mes.chat.id,
        "Вы действительно хотите удалить все ваши аккаунты? Это действие "
        "нельзя отменить. "
        "Отправьте YES для подтверждения",
    )
    register_state(mes, functools.partial(_delete_all2, bot, engine, bot_mes))


async def _delete_all2(
    bot: AsyncTeleBot, engine: Engine, prev_mes: Message, mes: Message
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "YES":
        database.delete.purge_accounts(engine, mes.from_user.id)
        database.delete.delete_master_pass(engine, mes.from_user.id)
        await send_tmp_message(
            bot,
            mes.chat.id,
            "Всё успешно удалено",
            sleep_time=10,
        )
    else:
        await send_tmp_message(
            bot,
            mes.chat.id,
            "Вы отправили не YES, ничего не удалено",
        )


async def set_master_password(
    bot: AsyncTeleBot,
    engine: Engine,
    mes: Message,
) -> None:
    await base_handler(bot, mes, None)
    if database.get.get_master_pass(engine, mes.from_user.id) is not None:
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Мастер пароль уже существует",
        )
    bot_mes = await bot.send_message(mes.chat.id, "Отправьте мастер пароль")
    register_state(
        mes,
        functools.partial(_set_master_pass2, bot, engine, bot_mes),
    )


async def _set_master_pass2(
    bot: AsyncTeleBot, engine: Engine, prev_mes: Message, mes: Message
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    master_password = encryption.master_pass.encrypt_master_pass(
        mes.from_user.id,
        text,
    )
    database.add.add_master_pass(engine, master_password)

    await send_tmp_message(bot, mes.chat.id, "Успех")
    del mes, text
    gc.collect()


async def reset_master_pass(
    bot: AsyncTeleBot,
    engine: Engine,
    mes: Message,
) -> None:
    await base_handler(bot, mes)

    if database.get.get_master_pass(engine, mes.from_user.id) is None:
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Мастер пароль не задан",
        )

    bot_mes = await bot.send_message(
        mes.chat.id,
        "Отправьте новый мастер пароль, осторожно, все текущие аккаунты "
        "будут удалены навсегда",
    )

    register_state(
        mes,
        functools.partial(_reset_master_pass2, bot, engine, bot_mes),
    )


async def _reset_master_pass2(
    bot: AsyncTeleBot, engine: Engine, prev_mes: Message, mes: Message
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    master_password = encryption.master_pass.encrypt_master_pass(
        mes.from_user.id,
        text,
    )
    database.delete.purge_accounts(engine, mes.from_user.id)
    database.change.change_master_pass(engine, master_password)

    await send_tmp_message(
        bot, mes.chat.id, "Все ваши аккаунты удалены, а мастер пароль изменён"
    )
    del mes, text
    gc.collect()


async def add_account(bot: AsyncTeleBot, engine: Engine, mes: Message) -> None:
    await base_handler(bot, mes)

    master_password_from_db = database.get.get_master_pass(
        engine,
        mes.from_user.id,
    )
    if master_password_from_db is None:
        return await send_tmp_message(bot, mes.chat.id, "Нет мастер пароля")

    bot_mes = await bot.send_message(
        mes.chat.id,
        "Отправьте название аккаунта",
    )

    register_state(
        mes,
        functools.partial(_add_account2, bot, engine, bot_mes),
    )


async def _add_account2(
    bot: AsyncTeleBot, engine: Engine, prev_mes: Message, mes: Message
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    if not check_account_name(text):
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Не корректное название аккаунта",
        )
    if text in database.get.get_accounts(engine, mes.from_user.id):
        return await send_tmp_message(
            bot, mes.chat.id, "Аккаунт с таким именем уже существует"
        )

    bot_mes = await bot.send_message(mes.chat.id, "Отправьте логин")
    data = {"name": text}

    register_state(
        mes,
        functools.partial(_add_account3, bot, engine, bot_mes, data),
    )


async def _add_account3(
    bot: AsyncTeleBot,
    engine: Engine,
    prev_mes: Message,
    data: dict[str, str],
    mes: Message,
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")
    if not check_login(text):
        return await send_tmp_message(bot, mes.chat.id, "Не корректный логин")

    data["login"] = text

    bot_mes = await bot.send_message(
        mes.chat.id,
        "Отправьте пароль от аккаунта",
    )

    register_state(
        mes,
        functools.partial(_add_account4, bot, engine, bot_mes, data),
    )


async def _add_account4(
    bot: AsyncTeleBot,
    engine: Engine,
    prev_mes: Message,
    data: dict[str, str],
    mes: Message,
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")
    if not check_password(text):
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Не корректный пароль",
        )

    data["passwd"] = text

    bot_mes = await bot.send_message(mes.chat.id, "Отправьте мастер пароль")

    register_state(
        mes,
        functools.partial(_add_account5, bot, engine, bot_mes, data),
    )


async def _add_account5(
    bot: AsyncTeleBot,
    engine: Engine,
    prev_mes: Message,
    data: dict[str, str],
    mes: Message,
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    master_password = database.get.get_master_pass(engine, mes.from_user.id)
    if not encryption.master_pass.check_master_pass(text, master_password):
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Не подходит главный пароль",
        )

    # name, login, passwd = data["name"], data["login"], data["passwd"]
    account = DecryptedAccount(
        user_id=mes.from_user.id,
        name=data["name"],
        login=data["login"],
        password=data["passwd"],
    )

    encrypted_account = encryption.other_accounts.encrypt(
        account,
        text,
    )

    result = database.add.add_account(
        engine,
        encrypted_account,
    )

    await send_tmp_message(
        bot,
        mes.chat.id,
        "Успех" if result else "Произошла не предвиденная ошибка",
    )

    del data, account

    gc.collect()


async def get_account(bot: AsyncTeleBot, engine: Engine, mes: Message) -> None:
    await base_handler(bot, mes)

    master_pass = database.get.get_master_pass(engine, mes.from_user.id)
    if master_pass is None:
        return await send_tmp_message(bot, mes.chat.id, "Нет мастер пароля")

    bot_mes = await bot.send_message(
        mes.chat.id,
        "Отправьте название аккаунта",
    )
    register_state(
        mes,
        functools.partial(_get_account2, bot, engine, bot_mes),
    )


async def _get_account2(
    bot: AsyncTeleBot, engine: Engine, prev_mes: Message, mes: Message
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    if text not in database.get.get_accounts(engine, mes.from_user.id):
        return await send_tmp_message(bot, mes.chat.id, "Нет такого аккаунта")

    bot_mes = await bot.send_message(mes.chat.id, "Отправьте мастер пароль")
    register_state(
        mes,
        functools.partial(_get_account3, bot, engine, bot_mes, text),
    )


async def _get_account3(
    bot: AsyncTeleBot,
    engine: Engine,
    prev_mes: Message,
    name: str,
    mes: Message,
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    master_password = database.get.get_master_pass(
        engine,
        mes.from_user.id,
    )

    if not encryption.master_pass.check_master_pass(text, master_password):
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Не подходит мастер пароль",
        )

    account = database.get.get_account_info(engine, mes.from_user.id, name)
    account = encryption.other_accounts.decrypt(
        account,
        text,
    )
    await send_deleteable_message(
        bot,
        mes.chat.id,
        f"Название:\n{escape(account.name)}\n"
        f"Логин:\n`{account.login}`\nПароль:\n`{account.password}`\nНажмите "
        "на логин или пароль, чтобы скопировать",
    )

    del text, mes
    gc.collect()


async def delete_account(
    bot: AsyncTeleBot,
    engine: Engine,
    mes: Message,
) -> None:
    await base_handler(bot, mes)

    master_pass = database.get.get_master_pass(engine, mes.from_user.id)
    if master_pass is None:
        return await send_tmp_message(bot, mes.chat.id, "Нет мастер пароля")

    bot_mes = await bot.send_message(
        mes.chat.id, "Отправьте название аккаунта, который вы хотите удалить"
    )

    register_state(
        mes,
        functools.partial(_delete_account2, bot, engine, bot_mes),
    )


async def _delete_account2(
    bot: AsyncTeleBot,
    engine: Engine,
    prev_mes: Message,
    mes: Message,
):
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    if text not in database.get.get_accounts(engine, mes.from_user.id):
        return await send_tmp_message(bot, mes.chat.id, "Нет такого аккаунта")

    bot_mes = await bot.send_message(
        mes.from_user.id,
        f'Вы уверены, что хотите удалить аккаунт "{text}"?\nОтправьте YES для '
        "подтверждения",
    )

    register_state(
        mes,
        functools.partial(_delete_account3, bot, engine, bot_mes, text),
    )


async def _delete_account3(
    bot: AsyncTeleBot,
    engine: Engine,
    prev_mes: Message,
    account_name: str,
    mes: Message,
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text != "YES":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    database.delete.delete_account(engine, mes.from_user.id, account_name)
    await send_tmp_message(bot, mes.chat.id, "Аккаунт удалён")


async def help_command(bot: AsyncTeleBot, mes: Message) -> None:
    message = """Команды:
/set_master_pass - установить мастер пароль
/add_account - создать аккаунт
/get_accounts - получить список аккаунтов
/get_account - получить логин и пароль аккаунта
/delete_account - удалить аккаунт
/delete_all - удалить все аккаунты и мастер пароль
/reset_master_pass - удалить все аккаунты и изменить мастер пароль
/cancel - отмена текущего действия
/help - помощь
/export - получить пароли в json формате
/import - импортировать пароли из json в файле в таком же формате, \
как из /export
/gen_password - создать 10 надёжных паролей"""
    await bot.send_message(mes.chat.id, message)


async def cancel(bot: AsyncTeleBot, mes: Message) -> None:
    await base_handler(bot, mes)
    await send_tmp_message(bot, mes.chat.id, "Нет активного действия")


async def export(bot: AsyncTeleBot, engine: Engine, mes: Message) -> None:
    await base_handler(bot, mes)
    master_password_from_db = database.get.get_master_pass(
        engine,
        mes.from_user.id,
    )

    if master_password_from_db is None:
        return await send_tmp_message(bot, mes.chat.id, "Нет мастер пароля")

    if not database.get.get_accounts(engine, mes.from_user.id):
        return await send_tmp_message(bot, mes.chat.id, "Нет аккаунтов")

    bot_mes = await bot.send_message(mes.chat.id, "Отправьте мастер пароль")

    register_state(mes, functools.partial(_export2, bot, engine, bot_mes))


async def _export2(
    bot: AsyncTeleBot, engine: Engine, prev_mes: Message, mes: Message
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    master_password = database.get.get_master_pass(
        engine,
        mes.from_user.id,
    )
    if not encryption.master_pass.check_master_pass(text, master_password):
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Не подходит мастер пароль",
        )

    accounts = database.get.get_all_accounts(engine, mes.from_user.id)
    accounts = encryption.other_accounts.decrypt_multiple(accounts, text)
    json_io = accounts_to_json(accounts)
    await bot.send_document(
        mes.chat.id,
        json_io,
        reply_markup=markups.deletion_markup(),
    )

    del text, accounts, json_io
    gc.collect()


async def import_accounts(
    bot: AsyncTeleBot,
    engine: Engine,
    mes: Message,
) -> None:
    await base_handler(bot, mes)
    master_password_from_db = database.get.get_master_pass(
        engine,
        mes.from_user.id,
    )

    if master_password_from_db is None:
        return await send_tmp_message(bot, mes.chat.id, "Нет мастер пароля")

    bot_mes = await bot.send_message(mes.chat.id, "Отправьте json файл")

    register_state(mes, functools.partial(_import2, bot, engine, bot_mes))


async def _import2(
    bot: AsyncTeleBot, engine: Engine, prev_mes: Message, mes: Message
) -> None:
    await base_handler(bot, mes, prev_mes)
    if mes.text is not None:
        text = mes.text.strip()
        if text == "/cancel":
            return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    if mes.document is None:
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Вы должны отправить документ",
        )
    if mes.document.file_size > 102_400:  # If file size is bigger than 100 MB
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Файл слишком большой",
        )

    file_info = await bot.get_file(mes.document.file_id)
    downloaded_file = await bot.download_file(file_info.file_path)
    try:
        accounts = json_to_accounts(
            downloaded_file.decode("utf-8"),
            mes.from_user.id,
        )
    except Exception:
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Ошибка во время работы с файлом",
        )

    bot_mes = await bot.send_message(mes.chat.id, "Отправьте мастер пароль")
    register_state(
        mes,
        functools.partial(_import3, bot, engine, bot_mes, accounts),
    )


async def _import3(
    bot: AsyncTeleBot,
    engine: Engine,
    prev_mes: Message,
    accounts: list[DecryptedAccount],
    mes: Message,
) -> None:
    await base_handler(bot, mes, prev_mes)
    text = mes.text.strip()
    if text == "/cancel":
        return await send_tmp_message(bot, mes.chat.id, "Успешная отмена")

    master_password = database.get.get_master_pass(
        engine,
        mes.from_user.id,
    )
    if not encryption.master_pass.check_master_pass(text, master_password):
        return await send_tmp_message(
            bot,
            mes.chat.id,
            "Не подходит мастер пароль",
        )

    # List of names of accounts, which failed to be added to the database
    # or failed the tests
    failed: list[str] = []
    for account in accounts:
        if not check_account(account):
            failed.append(account.name)
            continue
        account = encryption.other_accounts.encrypt(
            account,
            text,
        )
        result = database.add.add_account(engine, account)
        if not result:
            failed.append(account.name)

    if failed:
        mes_text = "Не удалось добавить:\n" + "\n".join(failed)
    else:
        mes_text = "Успех"

    await send_tmp_message(bot, mes.chat.id, mes_text, 10)
    del text, mes, accounts
    gc.collect()


async def gen_password(bot: AsyncTeleBot, mes: Message) -> None:
    await base_handler(bot, mes)
    # Generate 10 passwords and put 'em in the backticks
    passwords = (f"`{generate_password.gen_password()}`" for _ in range(10))
    text = (
        "Пароли:\n"
        + "\n".join(passwords)
        + "\nНажмите на пароль, чтобы его скопировать"
    )
    await send_tmp_message(bot, mes.chat.id, text, 15)


async def message_handler(bot: AsyncTeleBot, mes: Message) -> None:
    handler = get_state(mes)
    if handler is None:
        await delete_message(bot, mes)
        await send_tmp_message(
            bot,
            mes.chat.id,
            "Вы отправили не корректное сообщение",
        )
        return
    await handler(mes)
