import base64
import os
from typing import Iterable, Iterator

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from ..database.models import Account
from ..decrypted_account import DecryptedAccount


def _generate_key(salt: bytes, master_pass: bytes) -> bytes:
    """Generates key for fernet encryption"""
    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=salt,
        iterations=100000,
        backend=default_backend(),
    )
    key = base64.urlsafe_b64encode(kdf.derive(master_pass))
    return key


def encrypt(
    account: DecryptedAccount,
    master_pass: str,
) -> Account:
    """Encrypts account using master password and returns Account object"""
    salt = os.urandom(64)
    key = _generate_key(salt, master_pass.encode("utf-8"))
    f = Fernet(key)

    enc_login = base64.urlsafe_b64decode(
        f.encrypt(account.login.encode("utf-8")),
    )
    enc_password = base64.urlsafe_b64decode(
        f.encrypt(account.password.encode("utf-8")),
    )

    return Account(
        user_id=account.user_id,
        name=account.name,
        salt=salt,
        enc_login=enc_login,
        enc_password=enc_password,
    )


def decrypt(
    account: Account,
    master_pass: str,
) -> DecryptedAccount:
    """Decrypts account using master password and returns
    DecryptedAccount object"""
    key = _generate_key(account.salt, master_pass.encode("utf-8"))
    f = Fernet(key)

    login = f.decrypt(
        base64.urlsafe_b64encode(account.enc_login),
    ).decode("utf-8")
    password = f.decrypt(
        base64.urlsafe_b64encode(account.enc_password),
    ).decode("utf-8")

    return DecryptedAccount(
        user_id=account.user_id,
        name=account.name,
        login=login,
        password=password,
    )


def decrypt_multiple(
    accounts: Iterable[Account], master_pass: str
) -> Iterator[DecryptedAccount]:
    """Decrypts an iterable of accounts using master_pass and
    returns an Iterator of decrypted accounts"""
    for account in accounts:
        yield decrypt(account, master_pass)
